import qiskit
import numpy as np
import pennylane as qml

from .base import BaseGate


class GlobalPhase(BaseGate):
    """
    Class representing a Global Phase
    :param phase: Global phase angle in radians
    """

    def __new__(cls, phase: float) -> "GlobalPhase":
        input_array = np.array([[np.exp(1j * phase)]], dtype=complex)
        obj = super().__new__(cls, input_array)
        assert obj.is_1x1(), "Quscalar must be a 1x1 matrix"
        return obj

    @property
    def scalar(self) -> complex:
        """Get the scalar value of the gate."""
        return self[0, 0]

    @property
    def phase(self) -> float:
        """Get the global phase."""
        return np.angle(self.scalar)

    def is_1x1(self) -> bool:
        """Check if the gate is a 1x1 matrix."""
        return self.shape == (1, 1)

    def inverse(self) -> "GlobalPhase":
        inverse_phase = -self.phase
        return GlobalPhase(inverse_phase)

    def combine(self, other: "GlobalPhase") -> "GlobalPhase":
        assert isinstance(other, GlobalPhase), "Can only combine with another QuScalarGate."
        combined_phase = self.phase + other.phase
        return GlobalPhase(combined_phase)

    def multiply(self, other: "GlobalPhase") -> "GlobalPhase":
        assert isinstance(other, GlobalPhase), "Can only multiply with another QuScalarGate."
        multiplied_phase = np.angle(self.scalar * other.scalar)
        return GlobalPhase(multiplied_phase)

    def to_qiskit(self) -> qiskit.circuit.library.GlobalPhaseGate:
        """Convert QuScalar to a Qiskit GlobalPhaseGate object."""
        return qiskit.circuit.library.GlobalPhaseGate(self.phase)

    @staticmethod
    def from_qiskit(qiskit_gate: qiskit.circuit.library.GlobalPhaseGate) -> "GlobalPhase":
        """
        Convert a Qiskit GlobalPhaseGate to a QuScalar object.
        :param qiskit_gate: Qiskit GlobalPhaseGate object
        :return: A QuScalar object
        """
        if not isinstance(qiskit_gate, qiskit.circuit.library.GlobalPhaseGate):
            raise ValueError(f"Expected GlobalPhaseGate, got {type(qiskit_gate)}.")
        phase = qiskit_gate.params[0]
        return GlobalPhase(phase)

    def to_pennylane(self) -> qml.GlobalPhase:
        """Convert QuScalar to a PennyLane GlobalPhase."""
        return qml.GlobalPhase(self.phase)

    @staticmethod
    def from_pennylane(pennylane_gate: qml.operation.Operation) -> "GlobalPhase":
        """
        Convert a PennyLane GlobalPhase to a QuScalar object.
        :param pennylane_gate: PennyLane GlobalPhase object
        :return: A QuScalar object
        """
        if not isinstance(pennylane_gate, qml.GlobalPhase):
            raise ValueError(f"Expected GlobalPhase, got {type(pennylane_gate)}.")
        phase = pennylane_gate.parameters[0]
        return GlobalPhase(phase)


class Identity(GlobalPhase):
    """No phase shift."""

    def __new__(cls) -> "Identity":
        return super().__new__(cls, 0.0)


class QuarterPhase(GlobalPhase):
    """Quarter phase shift (π/2)"""

    def __new__(cls) -> "QuarterPhase":
        return super().__new__(cls, np.pi / 2)


class HalfPhase(GlobalPhase):
    """Half phase shift (π)"""

    def __new__(cls) -> "HalfPhase":
        return super().__new__(cls, np.pi)


class ThreeQuarterPhase(GlobalPhase):
    """Three quarters phase shift (3π/2)"""

    def __new__(cls) -> "ThreeQuarterPhase":
        return super().__new__(cls, 3 * np.pi / 2)


class FullPhase(GlobalPhase):
    """Full phase shift (2π)"""

    def __new__(cls) -> "FullPhase":
        return super().__new__(cls, 2 * np.pi)



import numpy as np
from plotly import graph_objects as go


def _check_quantum_state_array(X):
    """Ensure that input are normalized state vectors of the right size."""
    if len(X.shape) != 2:
        raise ValueError("Input must be a 2D array.")

    # Check if all inputs are complex
    if not np.iscomplexobj(X):
        raise ValueError("Input must be a complex array.")

    # Check if input is normalized.
    normalized = np.allclose(np.linalg.norm(X, axis=-1), 1)
    if not normalized:
        not_normalized = X[np.linalg.norm(X, axis=-1) != 1]
        raise ValueError(f"Input state vectors must be normalized. Got non-normalized vectors: '{not_normalized}'")

    # Check if input is a valid state vector
    _, n_states = X.shape
    num_qubits = int(np.log2(n_states))
    assert n_states == 2**num_qubits, f"Expected state vector length {2**num_qubits}, got {n_states}."
    return True


def to_bitstring(arr: np.array) -> list[str]:
    """
    Convert a 2D binary array to a list of bitstrings.
    :param arr: A 2D binary array
    For example:
    [[0, 1],
     [1, 0]]
    :return: List of bitstrings
    For example:
    ['01', '10']
    """
    return ["".join(map(str, m)) for m in arr]


def matrix_heatmap(matrix: np.array, title: str = "Matrix Heatmap") -> go.Figure:
    """
    Visualize a matrix as a heatmap.
    :param matrix: A 2D array
    :param title: Title of the plot
    :return: A plotly figure
    """
    fig = go.Figure(data=go.Heatmap(z=np.real(matrix), colorscale="RdBu", zmin=-1, zmax=1, text=np.real(matrix).round(3), texttemplate="%{text}", textfont={"size": 10}, hoverongaps=False))
    fig.update_layout(title=title, xaxis_title="Column Index", yaxis_title="Row Index")
    return fig


def bloch(state: np.array) -> go.Figure:
    """
    Plot single qubit state on a Bloch sphere (3D plot).
    :param state: A statevector array for a single qubit
    :return: A plotly figure
    """

    def calculate_coordinates(theta, phi):
        x = np.sin(theta) * np.cos(phi)
        y = np.sin(theta) * np.sin(phi)
        z = np.cos(theta)
        return x, y, z

    fig = go.Figure()
    alpha, beta = state[0], state[1]
    theta = 2 * np.arccos(np.abs(alpha))
    phi = np.angle(beta) - np.angle(alpha)
    x, y, z = calculate_coordinates(theta, phi)
    surface_phi, surface_theta = np.mgrid[0 : 2 * np.pi : 100j, 0 : np.pi : 50j]
    xs, ys, zs = calculate_coordinates(surface_theta, surface_phi)

    fig.add_trace(go.Surface(x=xs, y=ys, z=zs, opacity=0.5, colorscale="Blues", showscale=False))

    fig.add_trace(go.Scatter3d(x=[0, x], y=[0, y], z=[0, z], mode="lines+markers+text", marker=dict(size=10, color="green"), line=dict(color="green", width=8), textposition="top center", showlegend=True, name=f"{alpha:.3f}|0⟩ + {beta:.3f}|1⟩"))

    fig.add_trace(go.Scatter3d(x=[0, 0, 1, -1, 0, 0], y=[0, 0, 0, 0, 1, -1], z=[1, -1, 0, 0, 0, 0], mode="markers", marker=dict(size=5, color="black"), hovertext=["|0⟩", "|1⟩", "|+⟩", "|-⟩", "|i⟩", "|-i⟩"], showlegend=False, name="Basis states"))

    boundary_phi = np.linspace(0, 2 * np.pi, 100)
    coords = [(np.cos(boundary_phi), np.sin(boundary_phi), np.zeros_like(boundary_phi)), (np.zeros_like(boundary_phi), np.cos(boundary_phi), np.sin(boundary_phi)), (np.cos(boundary_phi), np.zeros_like(boundary_phi), np.sin(boundary_phi))]

    for x, y, z in coords:
        fig.add_trace(go.Scatter3d(x=x, y=y, z=z, mode="lines", line=dict(color="black", width=2), showlegend=False, name="Axes"))

    fig.update_layout(
        legend=dict(
            font=dict(size=20),
            x=0.05,
            y=0.95,
            xanchor="left",
            yanchor="top",
            bgcolor="rgba(0,0,0,0)",
        ),
        margin=dict(l=0, r=0, t=0, b=0),
    )
    return fig


def measure_state(state: np.array) -> int:
    """
    Measure a quantum state and return a classical bit.
    :param state: A statevector array
    :return: A classical bit
    """
    # Absolute value squared gives us the probability distribution from the statevector
    p = np.abs(state) ** 2
    # Sample from probability distribution to get classical bits
    return np.random.choice([0, 1], p=p)



import numpy as np
from ..base import Operator


class BaseGate(Operator):
    """
    Base class for quantum gates of any computational basis.
    The gate must be a 2D unitary matrix.
    :param input_array: Input array to create the gate. Will be converted to a NumPy array.
    """

    def __new__(cls, input_array) -> "BaseGate":
        obj = super().__new__(cls, input_array)
        assert obj.is_unitary(), "Gate must be unitary."
        return obj

    def is_unitary(self) -> bool:
        """Check if the gate is unitary (U*U^dagger = I)"""
        identity = np.eye(self.shape[0])
        return np.allclose(self @ self.conjugate_transpose(), identity)

    def kernel_density(self, other: "BaseGate") -> complex:
        """
        Calculate the quantum kernel using density matrices.
        The kernel is defined as Tr(U * V).
        :param other: Gate object to compute the kernel with.
        :return kernel: Complex number representing the kernel density.
        """
        assert isinstance(other, BaseGate), "Other object must be an instance of BaseUnitary (i.e. a Unitary matrix)."
        assert self.num_levels() == other.num_levels(), "Gates must have the same number of rows for the kernel density."
        return np.trace(self @ other)

    def hilbert_schmidt_inner_product(self, other: "BaseGate") -> complex:
        """
        Calculate the Hilbert-Schmidt inner product with another gate.
        The inner product is Tr(U^dagger * V).
        :param other: Gate object to compute the inner product with.
        :return inner_product: Complex number representing the inner product.
        """
        assert isinstance(other, BaseGate), "Other object must be an instance of BaseUnitary (i.e. a Unitary matrix)."
        assert self.num_levels() == other.num_levels(), "Gates must have the same number of rows for the Hilbert-Schmidt inner product."
        return np.trace(self.conjugate_transpose() @ other)

    def convert_endianness(self) -> "BaseGate":
        """Convert a gate matrix from big-endian to little-endian and vice versa."""
        num_qubits = self.num_qubits
        perm = np.argsort([int(bin(i)[2:].zfill(num_qubits)[::-1], 2) for i in range(2**num_qubits)])
        return self[np.ix_(perm, perm)]



from typing import Callable

from ..gates.qubit.multi import DeutschOracle, DeutschJozsaOracle
from ..gates.qubit.single import I, X, H, Measure
from .circuit import Circuit, Concat


class Deutsch:
    """Deutsch's algorithm."""

    def circuit(self, f: Callable, measure: bool = True) -> Circuit:
        """
        Deutsch's algorithm
        :param f: Binary function that maps a single bit to a single bit.
        :param measure: Whether to measure the qubits.
        :return: skq Circuit that implements Deutsch's algorithm.
        """
        circuit = [
            Concat([I(), X()]),
            Concat([H(), H()]),
            DeutschOracle(f),
            Concat([H(), I()]),
        ]
        if measure:
            circuit.append(Measure())
        return Circuit(circuit)


class DeutschJozsa:
    """Deutsch-Jozsa algorithm."""

    def circuit(self, f: Callable, n_bits: int, measure: bool = True) -> Circuit:
        """
        Deutsch-Josza algorithm
        :param f: Binary function that maps a single bit to a single bit.
        :return: skq Circuit that implements Deutsch-Josza algorithm.
        """
        circuit = [
            Concat([I() for _ in range(n_bits - 1)] + [X()]),
            Concat([H() for _ in range(n_bits)]),
            DeutschJozsaOracle(f, n_bits=n_bits),
            Concat([H() for _ in range(n_bits)]),
        ]
        if measure:
            circuit.append(Measure())
        return Circuit(circuit)



import numpy as np

# Planck constants in J s
PLANCK_CONSTANT = 6.62607015e-34
REDUCED_PLANCK_CONSTANT = 1.054571817e-34
# Boltzmann constant in J/K
BOLTZMANN_CONSTANT = 1.380649e-23
# Speed of light in m/s
SPEED_OF_LIGHT = 2.99792458e8
# Electron charge in C
ELECTRON_CHARGE = 1.602176634e-19
# Permeability of free space in N/A^2
PERMEABILITY_OF_FREE_SPACE = 4 * np.pi * 1e-7
# Permittivity of free space in F/m
PERMITTIVITY_OF_FREE_SPACE = 8.854187817e-12



import numpy as np

from ..gates.qubit.multi import PhaseOracle, GroverDiffusion
from ..gates.qubit.single import H, Measure
from .circuit import Circuit, Concat


class Grover:
    """
    Grover's search algorithm.
    """

    def circuit(self, target_state: np.array, n_qubits: int, n_iterations: int, measure: bool = True) -> Circuit:
        """
        Grover's search algorithm
        :param target_state: Target state to search for.
        :param n_qubits: Number of qubits in the circuit.
        :param n_iterations: Number of Grover iterations to perform.
        :param measure: Whether to measure the qubits.
        :return: Circuit for the Grover search algorithm.
        """
        single_grover_iteration = [PhaseOracle(target_state), GroverDiffusion(n_qubits=n_qubits)]
        equal_superposition = Concat([H() for _ in range(n_qubits)])
        steps = [equal_superposition, *[gate for _ in range(n_iterations) for gate in single_grover_iteration]]
        if measure:
            steps.append(Measure())
        return Circuit(steps)



import numpy as np
from ..base import BaseGate


class QuquartGate(BaseGate):
    """
    Base class for Ququart gates.
    These are quantum systems with a basis of 4 states. |0>, |1>, |2>, |3>.
    Models spin-1/2 particles like electrons.
    """

    def __new__(cls, input_array):
        obj = super().__new__(cls, input_array)
        assert obj.is_at_least_nxn(n=4), "Gate must be at least a 4x4 matrix"
        assert obj.is_power_of_n_shape(n=4), "Gate shape must be a power of 4"
        return obj

    def num_qudits(self) -> int:
        """Return the number of qudits involved in the gate."""
        return int(np.log(self.shape[0]) / np.log(4))

    def is_multi_qudit(self) -> bool:
        """Check if the gate involves multiple qudits."""
        return self.num_qudits() > 1



import numpy as np
from qiskit import QuantumCircuit
from functools import reduce


from ..base import Operator
from ..gates.qubit import I, Measure


class Circuit(list):
    """Run multiple qubit gates in sequence."""

    @property
    def num_qubits(self) -> int:
        return max(g.num_qubits for g in self)

    def encodes(self, x):
        for gate in self:
            x = gate.encodes(x)
        return x

    def decodes(self, x):
        for gate in reversed(self):
            x = gate.decodes(x)
        return x

    def __call__(self, x):
        return self.encodes(x)

    def convert(self, framework="qiskit"):
        """Convert circuit to a given framework.
        :param framework: Framework to convert to.
        :return: Converter Circuit object.
        For Qiskit -> QuantumCircuit object.
        For QASM -> OpenQASM string.
        """
        converter_mapping = {"qiskit": convert_to_qiskit, "qasm": convert_to_qasm}
        assert framework in converter_mapping, f"Invalid framework. Supported frameworks: {converter_mapping.keys()}."
        return converter_mapping[framework](self)

    def draw(self, **kwargs):
        """Draw circuit using Qiskit."""
        return self.convert(framework="qiskit").draw(**kwargs)


class Concat:
    """
    Combine multiple qubit gates into a single gate.
    :param gates: List of gates to concatenate.
    """

    def __init__(self, gates: list[Operator]):
        assert len(gates) > 1, "Concat must have at least 2 gates."
        assert all(isinstance(g, Operator) for g in gates), "All gates must be instances of Operator."
        self.gates = gates
        self.encoding_matrix = reduce(np.kron, [g for g in gates])
        self.num_qubits = sum(g.num_qubits for g in gates)

    def encodes(self, x: np.ndarray) -> np.ndarray:
        """
        Concatenate 2 or more gates.

        :param x: Quantum state to encode.
        :return: Quantum state after encoding.
        """
        return x @ self.encoding_matrix

    def decodes(self, x: np.ndarray) -> np.ndarray:
        """
        Reverse propagation for all gates.

        :param x: Quantum state to decode.
        :return: Quantum state after decoding.
        """
        for g in reversed(self.gates):
            x = x @ np.kron(g.conj().T, np.eye(len(x) // g.shape[0]))
        return x

    def __call__(self, x):
        return self.encodes(x)


def convert_to_qiskit(circuit: Circuit) -> QuantumCircuit:
    """Convert a skq Circuit into a Qiskit QuantumCircuit."""

    def handle_gate(qc, gate, start_idx=None):
        if isinstance(gate, I):
            return
        if isinstance(gate, Measure):
            for q in range(circuit.num_qubits):
                qc.measure(q, q)
            return
        # Create list of consecutive qubit indices starting from start_idx
        qubits = list(range(start_idx, start_idx + gate.num_qubits)) if start_idx is not None else list(range(gate.num_qubits))
        qc.append(gate.to_qiskit(), qubits)

    qc = QuantumCircuit(circuit.num_qubits, circuit.num_qubits if any(isinstance(g, Measure) for g in circuit) else 0)

    for gate in circuit:
        if isinstance(gate, Concat):
            offset = 0
            for sub_gate in gate.gates:
                handle_gate(qc, sub_gate, offset)
                offset += sub_gate.num_qubits
        else:
            handle_gate(qc, gate)
    return qc


def convert_to_qasm(circuit: Circuit) -> str:
    """Convert a skq Circuit into an OpenQASM string."""
    qasm_lines = [
        "OPENQASM 2.0;",  # OpenQASM header
        'include "qelib1.inc";',  # Standard library
        f"qreg q[{circuit.num_qubits}];",  # Declare quantum register
    ]
    # Declare classical register if measurement is present
    if any(isinstance(g, Measure) for g in circuit):
        qasm_lines.append(f"creg c[{circuit.num_qubits}];")

    def handle_gate(gate, start_idx=None):
        if isinstance(gate, I):
            return []
        if isinstance(gate, Measure):
            return [f"measure q[{q}] -> c[{q}];" for q in range(circuit.num_qubits)]
        # Create list of consecutive qubit indices starting from start_idx
        qubits = list(range(start_idx, start_idx + gate.num_qubits)) if start_idx is not None else list(range(gate.num_qubits))
        return [gate.to_qasm(qubits)]

    for gate in circuit:
        if isinstance(gate, Concat):
            offset = 0
            for sub_gate in gate.gates:
                qasm_lines.extend(handle_gate(sub_gate, offset))
                offset += sub_gate.num_qubits
        else:
            qasm_lines.extend(handle_gate(gate))
    return "\n".join(qasm_lines)



import numpy as np

from skq.constants import *


def photon_energy(frequency):
    return PLANCK_CONSTANT * frequency


def blackbody_radiance(frequency, temperature):
    return (2 * PLANCK_CONSTANT * frequency**3 / SPEED_OF_LIGHT**2) / (np.exp(PLANCK_CONSTANT * frequency / (BOLTZMANN_CONSTANT * temperature)) - 1)


def hydrogen_energy_level(n):
    return -(REDUCED_PLANCK_CONSTANT**2 * SPEED_OF_LIGHT * ELECTRON_CHARGE**4) / (8 * np.pi**2 * PERMITTIVITY_OF_FREE_SPACE**2 * REDUCED_PLANCK_CONSTANT**2 * n**2)


def coulomb_force(q1, q2, r):
    return (1 / (4 * np.pi * PERMITTIVITY_OF_FREE_SPACE)) * (q1 * q2) / r**2


def test_photon_energy():
    frequency = 1e14
    expected_energy = 6.62607015e-20
    calculated_energy = photon_energy(frequency)
    assert np.isclose(calculated_energy, expected_energy, rtol=1e-9), f"Photon energy is incorrect: {calculated_energy} != {expected_energy}"


def test_blackbody_radiance():
    frequency = 1e14
    temperature = 300
    expected_radiance = 3.74177e-19 / (np.exp(4.799243073e-02) - 1)
    calculated_radiance = blackbody_radiance(frequency, temperature)
    assert np.isclose(calculated_radiance, expected_radiance, rtol=1e-9), f"Blackbody radiance is incorrect: {calculated_radiance} != {expected_radiance}"


def test_permittivity_and_permeability_of_free_space():
    expected_speed_of_light = 1 / np.sqrt(PERMEABILITY_OF_FREE_SPACE * PERMITTIVITY_OF_FREE_SPACE)
    assert np.isclose(expected_speed_of_light, SPEED_OF_LIGHT, rtol=1e-9), f"Calculated speed of light is incorrect: {expected_speed_of_light} != {SPEED_OF_LIGHT}"


def test_hydrogen_energy_level():
    expected_energy = -2.1798723611035e-18
    calculated_energy = hydrogen_energy_level(1)
    assert np.isclose(calculated_energy, expected_energy, rtol=1e-9), f"Hydrogen energy level is incorrect: {calculated_energy} != {expected_energy}"


def test_coulomb_force():
    expected_force = 2.307077551e-8
    calculated_force = coulomb_force(ELECTRON_CHARGE, ELECTRON_CHARGE, 1e-10)
    assert np.isclose(calculated_force, expected_force, rtol=1e-9), f"Coulomb force is incorrect: {calculated_force} != {expected_force}"



import qiskit
import numpy as np
import pennylane as qml

from .base import QubitGate
from .single import X, Y, Z
from ...quantum_info.state import Statevector


class DeutschOracle(QubitGate):
    """
    Oracle for Deutsch algorithm with ancilla qubit.
    Implements |x,y⟩ -> |x, y⊕f(x)⟩

    :param f: Function that takes an integer x (0 or 1) and returns 0 or 1
    """

    def __new__(cls, f):
        matrix = np.zeros((4, 4))
        for x in [0, 1]:
            matrix[x * 2 + f(x), x * 2] = 1  # |x,0⟩ -> |x,f(x)⟩
            matrix[x * 2 + (1 - f(x)), x * 2 + 1] = 1  # |x,1⟩ -> |x,1-f(x)⟩

        return super().__new__(cls, matrix)

    def to_qiskit(self) -> qiskit.circuit.library.UnitaryGate:
        # Reverse the order of qubits for Qiskit's little-endian convention
        return qiskit.circuit.library.UnitaryGate(self.convert_endianness(), label="DeutschOracle")


class DeutschJozsaOracle(QubitGate):
    """
    Oracle for Deutsch-Jozsa algorithm.

    For input x and output f(x), the oracle applies phase (-1)^f(x).
    - Constant function: f(x) is same for all x
    - Balanced function: f(x) = 1 for exactly half of inputs

    :param f: Function that takes an integer x (0 to 2^n-1) and returns 0 or 1
    :param n_bits: Number of input bits
    """

    def __new__(cls, f, n_bits: int):
        n_states = 2**n_bits
        outputs = [f(x) for x in range(n_states)]
        if not all(v in [0, 1] for v in outputs):
            raise ValueError("Function must return 0 or 1")

        ones = sum(outputs)
        if ones != 0 and ones != n_states and ones != n_states // 2:
            raise ValueError("Function must be constant or balanced")

        oracle_matrix = np.eye(n_states)
        for x in range(n_states):
            oracle_matrix[x, x] = (-1) ** f(x)

        return super().__new__(cls, oracle_matrix)

    def to_qiskit(self) -> qiskit.circuit.library.UnitaryGate:
        # Reverse the order of qubits for Qiskit's little-endian convention
        return qiskit.circuit.library.UnitaryGate(self.convert_endianness(), label="DeutschJozsaOracle")


class PhaseOracle(QubitGate):
    """
    Phase Oracle as used in Grover's search algorithm.
    :param target_state: The target state to mark.
    target_state is assumed to be in Big-Endian format.
    """

    def __new__(cls, target_state: np.ndarray):
        state = Statevector(target_state)
        n_qubits = state.num_qubits
        identity = np.eye(2**n_qubits)
        oracle_matrix = identity - 2 * np.outer(target_state, target_state.conj())
        return super().__new__(cls, oracle_matrix)

    def to_qiskit(self) -> qiskit.circuit.library.UnitaryGate:
        # Reverse the order of qubits for Qiskit's little-endian convention
        return qiskit.circuit.library.UnitaryGate(self.convert_endianness(), label="PhaseOracle")


class GroverDiffusion(QubitGate):
    """
    Grover Diffusion Operator Gate as used in Grover's search algorithm.
    This gate amplifies the amplitude of the marked state.
    :param n_qubits: The number of qubits in the system.
    """

    def __new__(cls, n_qubits: int):
        assert n_qubits >= 1, "GroverDiffusionGate must have at least one qubit."
        # Equal superposition state vector
        size = 2**n_qubits
        equal_superposition = np.ones(size) / np.sqrt(size)
        # Grover diffusion operator: 2|ψ⟩⟨ψ| - I
        diffusion_matrix = 2 * np.outer(equal_superposition, equal_superposition) - np.eye(size)
        return super().__new__(cls, diffusion_matrix)

    def to_qiskit(self) -> qiskit.circuit.library.UnitaryGate:
        # Reverse the order of qubits for Qiskit's little-endian convention
        return qiskit.circuit.library.UnitaryGate(self.convert_endianness(), label="GroverDiffusion")


class QFT(QubitGate):
    """
    n-qubit Quantum Fourier Transform gate.
    :param n_qubits: The number of qubits in the system.
    :param inverse: Whether to use the inverse QFT.
    - For example, in Shor's algorithm we use the inverse QFT.
    """

    def __new__(cls, n_qubits: int, inverse: bool = False):
        dim = 2**n_qubits
        matrix = np.zeros((dim, dim), dtype=complex)
        # Use positive exponent for QFT, negative for QFT†
        sign = -1 if inverse else 1
        omega = np.exp(sign * 2j * np.pi / dim)

        for i in range(dim):
            for j in range(dim):
                matrix[i, j] = omega ** (i * j) / np.sqrt(dim)

        instance = super().__new__(cls, matrix)
        instance.inverse = inverse
        return instance

    def to_qiskit(self):
        """Convert to a Qiskit circuit implementing QFT or QFT†."""
        n = self.num_qubits
        name = "QFT†" if self.inverse else "QFT"

        def create_qft(n_qubits):
            circuit = qiskit.QuantumCircuit(n_qubits, name=name)
            for i in range(n_qubits):
                circuit.h(i)
                for j in range(i + 1, n_qubits):
                    circuit.cp(2 * np.pi / 2 ** (j - i + 1), i, j)
            return circuit

        qc = create_qft(n)
        if self.inverse:
            qc = qc.inverse()
        qc.name = name
        return qc

    def to_qasm(self, qubits: list[int]) -> str:
        """Convert to OpenQASM code."""
        n = self.num_qubits
        assert len(qubits) == n, f"OpenQASM QFT requires {n} qubits."

        name = "QFT†" if self.inverse else "QFT"
        qasm_str = f"// {name} circuit (big-endian convention)\n"

        if self.inverse:
            for i in range(n // 2):
                qasm_str += f"swap q[{qubits[i]}], q[{qubits[n-i-1]}];\n"
            for i in range(n - 1, -1, -1):
                for j in range(n - 1, i, -1):
                    angle = -np.pi / (2 ** (j - i))
                    qasm_str += f"cu1({angle}) q[{qubits[i]}], q[{qubits[j]}];\n"
                qasm_str += f"h q[{qubits[i]}];\n"
        else:
            for i in range(n):
                qasm_str += f"h q[{qubits[i]}];\n"
                for j in range(i + 1, n):
                    angle = np.pi / (2 ** (j - i))
                    qasm_str += f"cu1({angle}) q[{qubits[i]}], q[{qubits[j]}];\n"
            for i in range(n // 2):
                qasm_str += f"swap q[{qubits[i]}], q[{qubits[n-i-1]}];\n"

        qasm_str += f"// End of {name} circuit\n"
        return qasm_str


class CX(QubitGate):
    """
    Controlled-X (CNOT) gate.
    Used to entangle two qubits.
    If the control qubit is |1>, the target qubit is flipped.
    """

    def __new__(cls):
        return super().__new__(cls, [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]])

    def to_qiskit(self) -> qiskit.circuit.library.CXGate:
        return qiskit.circuit.library.CXGate()

    def to_pennylane(self, wires: list[int]) -> qml.CNOT:
        assert len(wires) == 2, "PennyLane CX gate requires 2 wires."
        return qml.CNOT(wires=wires)

    def to_qasm(self, qubits: list[int]) -> str:
        assert len(qubits) == 2, "OpenQASM CX gate requires 2 qubits."
        return f"cx q[{qubits[0]}], q[{qubits[1]}];"


class CY(QubitGate):
    """Controlled-Y gate."""

    def __new__(cls):
        return super().__new__(cls, [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, -1j], [0, 0, 1j, 0]])

    def to_qiskit(self) -> qiskit.circuit.library.CYGate:
        return qiskit.circuit.library.CYGate()

    def to_pennylane(self, wires: list[int]) -> qml.CY:
        assert len(wires) == 2, "PennyLane CY gate requires 2 wires."
        return qml.CY(wires=wires)

    def to_qasm(self, qubits: list[int]) -> str:
        assert len(qubits) == 2, "OpenQASM CY gate requires 2 qubits."
        return f"cy q[{qubits[0]}], q[{qubits[1]}];"


class CZ(QubitGate):
    """Controlled-Z gate."""

    def __new__(cls):
        return super().__new__(cls, [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, -1]])

    def to_qiskit(self) -> qiskit.circuit.library.CZGate:
        return qiskit.circuit.library.CZGate()

    def to_pennylane(self, wires: list[int]) -> qml.CZ:
        assert len(wires) == 2, "PennyLane CZ gate requires 2 wires."
        return qml.CZ(wires=wires)

    def to_qasm(self, qubits: list[int]) -> str:
        assert len(qubits) == 2, "OpenQASM CZ gate requires 2 qubits."
        return f"cz q[{qubits[0]}], q[{qubits[1]}];"


class CH(QubitGate):
    """Controlled-Hadamard gate."""

    def __new__(cls):
        return super().__new__(cls, [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1 / np.sqrt(2), 1 / np.sqrt(2)], [0, 0, 1 / np.sqrt(2), -1 / np.sqrt(2)]])

    def to_qiskit(self) -> qiskit.circuit.library.CHGate:
        return qiskit.circuit.library.CHGate()

    def to_pennylane(self, wires: list[int]) -> qml.CH:
        assert len(wires) == 2, "PennyLane CH gate requires 2 wires."
        return qml.CH(wires=wires)

    def to_qasm(self, qubits: list[int]) -> str:
        assert len(qubits) == 2, "OpenQASM CH gate requires 2 qubits."
        return f"ch q[{qubits[0]}], q[{qubits[1]}];"


class CPhase(QubitGate):
    """General controlled phase shift gate.
    :param phi: The phase shift angle in radians.
    """

    def __new__(cls, phi: float):
        obj = super().__new__(cls, [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, np.exp(1j * phi)]])
        obj.phi = phi
        return obj

    def to_qiskit(self) -> qiskit.circuit.library.CPhaseGate:
        return qiskit.circuit.library.CPhaseGate(self.phi)

    def to_pennylane(self, wires: list[int]) -> qml.CPhase:
        assert len(wires) == 2, "PennyLane CPhase gate requires 2 wires."
        return qml.CPhase(phi=self.phi, wires=wires)

    def to_qasm(self, qubits: list[int]) -> str:
        assert len(qubits) == 2, "OpenQASM CPhase gate requires 2 qubits."
        return f"cp({self.phi}) q[{qubits[0]}], q[{qubits[1]}];"


class CS(CPhase):
    """Controlled-S gate."""

    def __new__(cls):
        phi = np.pi / 2
        return super().__new__(cls, phi=phi)

    def to_qiskit(self) -> qiskit.circuit.library.CSGate:
        return qiskit.circuit.library.CSGate()


class CT(CPhase):
    """Controlled-T gate."""

    def __new__(cls):
        phi = np.pi / 4
        return super().__new__(cls, phi=phi)

    def to_qiskit(self) -> qiskit.circuit.library.TGate:
        return qiskit.circuit.library.TGate().control(1)


class SWAP(QubitGate):
    """Swap gate. Swaps the states of two qubits."""

    def __new__(cls):
        return super().__new__(cls, [[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]])

    def to_qiskit(self) -> qiskit.circuit.library.SwapGate:
        return qiskit.circuit.library.SwapGate()

    def to_pennylane(self, wires: list[int]) -> qml.SWAP:
        assert len(wires) == 2, "PennyLane SWAP gate requires 2 wires."
        return qml.SWAP(wires=wires)

    def to_qasm(self, qubits: list[int]) -> str:
        assert len(qubits) == 2, "OpenQASM SWAP gate requires 2 qubits."
        return f"swap q[{qubits[0]}], q[{qubits[1]}];"


class CSwap(QubitGate):
    """A controlled-SWAP gate. Also known as the Fredkin gate."""

    def __new__(cls):
        return super().__new__(cls, [[1, 0, 0, 0, 0, 0, 0, 0], [0, 1, 0, 0, 0, 0, 0, 0], [0, 0, 1, 0, 0, 0, 0, 0], [0, 0, 0, 1, 0, 0, 0, 0], [0, 0, 0, 0, 1, 0, 0, 0], [0, 0, 0, 0, 0, 0, 1, 0], [0, 0, 0, 0, 0, 1, 0, 0], [0, 0, 0, 0, 0, 0, 0, 1]])

    def to_qiskit(self) -> qiskit.circuit.library.CSwapGate:
        return qiskit.circuit.library.CSwapGate()

    def to_pennylane(self, wires: list[int]) -> qml.CSWAP:
        return qml.CSWAP(wires=wires)

    def to_qasm(self, qubits: list[int]) -> str:
        assert len(qubits) == 3, "OpenQASM CSWAP gate requires 3 qubits."
        return f"cswap q[{qubits[0]}], q[{qubits[1]}], q[{qubits[2]}];"


class CCX(QubitGate):
    """A 3-qubit controlled-controlled-X (CCX) gate. Also known as the Toffoli gate."""

    def __new__(cls):
        return super().__new__(cls, [[1, 0, 0, 0, 0, 0, 0, 0], [0, 1, 0, 0, 0, 0, 0, 0], [0, 0, 1, 0, 0, 0, 0, 0], [0, 0, 0, 1, 0, 0, 0, 0], [0, 0, 0, 0, 1, 0, 0, 0], [0, 0, 0, 0, 0, 1, 0, 0], [0, 0, 0, 0, 0, 0, 0, 1], [0, 0, 0, 0, 0, 0, 1, 0]])

    def to_qiskit(self) -> qiskit.circuit.library.CCXGate:
        return qiskit.circuit.library.CCXGate()

    def to_pennylane(self, wires: list[int]) -> qml.Toffoli:
        return qml.Toffoli(wires=wires)

    def to_qasm(self, qubits: list[int]) -> str:
        assert len(qubits) == 3, "OpenQASM CCX (Toffoli) gate requires 3 qubits."
        return f"ccx q[{qubits[0]}], q[{qubits[1]}], q[{qubits[2]}];"


class CCY(QubitGate):
    """A 3-qubit controlled-controlled-Y (CCY) gate."""

    def __new__(cls):
        return super().__new__(cls, [[1, 0, 0, 0, 0, 0, 0, 0], [0, 1, 0, 0, 0, 0, 0, 0], [0, 0, 1, 0, 0, 0, 0, 0], [0, 0, 0, 1, 0, 0, 0, 0], [0, 0, 0, 0, 1, 0, 0, 0], [0, 0, 0, 0, 0, 1, 0, 0], [0, 0, 0, 0, 0, 0, 0, -1j], [0, 0, 0, 0, 0, 0, 1j, 0]])

    def to_qiskit(self) -> qiskit.circuit.library.YGate:
        # There is no native CCY gate in Qiskit so we construct it.
        return qiskit.circuit.library.YGate().control(2)

    def to_qasm(self, qubits: list[int]) -> str:
        assert len(qubits) == 3, "OpenQASM CCY gate requires 3 qubits."
        # Custom OpenQASM implementation
        # sdg is an inverse s gate
        return f"""sdg q[{qubits[2]}];\ncx q[{qubits[1]}], q[{qubits[2]}];\ns q[{qubits[2]}];\ncx q[{qubits[0]}], q[{qubits[2]}];\nsdg q[{qubits[2]}];\ncx q[{qubits[1]}], q[{qubits[2]}];\ns q[{qubits[2]}];\ny q[{qubits[2]}];"""


class CCZ(QubitGate):
    """A 3-qubit controlled-controlled-Z (CCZ) gate."""

    def __new__(cls):
        return super().__new__(cls, [[1, 0, 0, 0, 0, 0, 0, 0], [0, 1, 0, 0, 0, 0, 0, 0], [0, 0, 1, 0, 0, 0, 0, 0], [0, 0, 0, 1, 0, 0, 0, 0], [0, 0, 0, 0, 1, 0, 0, 0], [0, 0, 0, 0, 0, 1, 0, 0], [0, 0, 0, 0, 0, 0, 1, 0], [0, 0, 0, 0, 0, 0, 0, -1]])

    def to_qiskit(self) -> qiskit.circuit.library.CCZGate:
        return qiskit.circuit.library.CCZGate()

    def to_qasm(self, qubits: list[int]) -> str:
        assert len(qubits) == 3, "OpenQASM CCZ gate requires 3 qubits."
        # CCZ = CCX sandwiched between two H gates on last qubit.
        return f"""h q[{qubits[2]}];\nccx q[{qubits[0]}], q[{qubits[1]}], q[{qubits[2]}];\nh q[{qubits[2]}];"""


class MCX(QubitGate):
    """
    Multi controlled-X (MCX) gate.
    :param num_ctrl_qubits: Number of control qubits.
    """

    def __new__(cls, num_ctrl_qubits: int):
        assert num_ctrl_qubits >= 1, "MCX gate must have at least one control qubit."
        cls.num_ctrl_qubits = num_ctrl_qubits
        levels = 2 ** (num_ctrl_qubits + 1)
        gate = np.identity(levels)
        gate[-2:, -2:] = X()
        return super().__new__(cls, gate)

    def to_qiskit(self) -> qiskit.circuit.library.CXGate | qiskit.circuit.library.CCXGate | qiskit.circuit.library.C3XGate | qiskit.circuit.library.C4XGate | qiskit.circuit.library.MCXGate:
        if self.num_ctrl_qubits == 1:
            return qiskit.circuit.library.CXGate()
        elif self.num_ctrl_qubits == 2:
            return qiskit.circuit.library.CCXGate()
        elif self.num_ctrl_qubits == 3:
            return qiskit.circuit.library.C3XGate()
        elif self.num_ctrl_qubits == 4:
            return qiskit.circuit.library.C4XGate()
        else:
            return qiskit.circuit.library.MCXGate(num_ctrl_qubits=self.num_ctrl_qubits)

    def to_pennylane(self, wires: list[int]) -> qml.CNOT | qml.QubitUnitary:
        if self.num_ctrl_qubits == 1:
            return qml.CNOT(wires=wires)
        else:
            return super().to_pennylane(wires)


class MCY(QubitGate):
    """Multi controlled-Y (MCY) gate."""

    def __new__(cls, num_ctrl_qubits: int):
        assert num_ctrl_qubits >= 1, "MCY gate must have at least one control qubit."
        cls.num_ctrl_qubits = num_ctrl_qubits
        levels = 2 ** (num_ctrl_qubits + 1)
        gate = np.identity(levels)
        gate[-2:, -2:] = Y()
        return super().__new__(cls, gate)

    def to_qiskit(self) -> qiskit.circuit.library.YGate:
        return qiskit.circuit.library.YGate().control(self.num_ctrl_qubits)

    def to_pennylane(self, wires: list[int]) -> qml.CY | qml.QubitUnitary:
        if self.num_ctrl_qubits == 1:
            return qml.CY(wires=wires)
        else:
            return super().to_pennylane(wires)


class MCZ(QubitGate):
    """Multi controlled-Z (MCZ) gate."""

    def __new__(cls, num_ctrl_qubits: int):
        assert num_ctrl_qubits >= 1, "MCZ gate must have at least one control qubit."
        cls.num_ctrl_qubits = num_ctrl_qubits
        levels = 2 ** (num_ctrl_qubits + 1)
        gate = np.identity(levels)
        gate[-2:, -2:] = Z()
        return super().__new__(cls, gate)

    def to_qiskit(self) -> qiskit.circuit.library.ZGate:
        return qiskit.circuit.library.ZGate().control(self.num_ctrl_qubits)

    def to_pennylane(self, wires: list[int]) -> qml.CZ | qml.QubitUnitary:
        if self.num_ctrl_qubits == 1:
            return qml.CZ(wires=wires)
        else:
            return super().to_pennylane(wires)


class CR(QubitGate):
    """Simple Cross-Resonance gate."""

    def __new__(cls, theta: float):
        return super().__new__(cls, [[1, 0, 0, 0], [0, np.cos(theta), 0, -1j * np.sin(theta)], [0, 0, 1, 0], [0, -1j * np.sin(theta), 0, np.cos(theta)]])


class SymmetricECR(QubitGate):
    """Symmetric Echoed Cross-Resonance gate."""

    def __new__(cls, theta: float):
        return super().__new__(cls, CR(theta) @ CR(-theta))


class AsymmetricECR(QubitGate):
    """Asymmetric Echoed Cross-Resonance gate."""

    def __new__(cls, theta1: float, theta2: float):
        return super().__new__(cls, CR(theta1) @ CR(theta2))


class CU(QubitGate):
    """
    General Controlled-Unitary gate.
    Applies a unitary operation conditionally based on a control qubit.
    If the control qubit is |1>, the unitary is applied to the target qubit(s).
    Only use this for custom unitaries. Else use standard controlled gates like CX, CY, CZ, CH, CS, CT, CPhase, etc.

    :param unitary: The unitary gate to be controlled. Must be a QubitGate.
    """

    def __new__(cls, unitary: QubitGate):
        assert isinstance(unitary, QubitGate), "Input must be a QubitGate"

        n_target_qubits = unitary.num_qubits
        dim = 2 ** (n_target_qubits + 1)
        matrix = np.eye(dim, dtype=complex)
        block_size = 2**n_target_qubits
        matrix[block_size:, block_size:] = unitary
        obj = super().__new__(cls, matrix)
        obj.unitary = unitary
        obj.n_target_qubits = n_target_qubits
        return obj

    def to_qiskit(self) -> qiskit.circuit.library.UnitaryGate:
        try:
            base_gate = self.unitary.to_qiskit()
            return base_gate.control(1)
        except (AttributeError, TypeError, ValueError):
            return qiskit.circuit.library.UnitaryGate(self.convert_endianness(), label=f"c-{self.unitary.__class__.__name__}")

    def to_pennylane(self, wires: list[int]) -> qml.QubitUnitary:
        """Convert to a PennyLane controlled gate."""
        assert len(wires) == self.num_qubits, f"PennyLane CU gate requires {self.num_qubits} wires."
        return qml.QubitUnitary(self, wires=wires)


class MCU(QubitGate):
    """
    Multi-Controlled Unitary gate.
    Applies a unitary operation conditionally based on multiple control qubits.
    The unitary is applied to target qubit(s) only if all control qubits are |1>.
    Only use this for custom unitaries. Else use standard controlled gates like CX, CY, CZ, CH, CS, CT, CPhase, etc.


    :param unitary: The unitary gate to be controlled. Must be a QubitGate.
    :param num_ctrl_qubits: Number of control qubits.
    """

    def __new__(cls, unitary: QubitGate, num_ctrl_qubits: int):
        assert isinstance(unitary, QubitGate), "Input must be a QubitGate"
        assert num_ctrl_qubits >= 1, "MCU gate must have at least one control qubit."
        n_target_qubits = unitary.num_qubits
        total_qubits = n_target_qubits + num_ctrl_qubits
        dim = 2**total_qubits
        matrix = np.eye(dim, dtype=complex)
        unitary_size = 2**n_target_qubits
        start_idx = dim - unitary_size
        matrix[start_idx:, start_idx:] = unitary

        obj = super().__new__(cls, matrix)
        obj.unitary = unitary
        obj.num_ctrl_qubits = num_ctrl_qubits
        obj.n_target_qubits = n_target_qubits
        return obj

    def to_qiskit(self) -> qiskit.circuit.library.UnitaryGate:
        """Convert to a Qiskit controlled gate."""
        try:
            base_gate = self.unitary.to_qiskit()
            return base_gate.control(self.num_ctrl_qubits)
        except (AttributeError, TypeError, ValueError):
            return qiskit.circuit.library.UnitaryGate(self.convert_endianness(), label=f"mc-{self.unitary.__class__.__name__}")

    def to_pennylane(self, wires: list[int]) -> qml.QubitUnitary:
        """Convert to a PennyLane multi-controlled gate."""
        assert len(wires) == self.num_qubits, f"PennyLane MCU gate requires {self.num_qubits} wires."
        return qml.QubitUnitary(self, wires=wires)



import numpy as np
from ..base import BaseGate


class QupentGate(BaseGate):
    """
    Base class for Qupent gates.
    These are quantum systems with a basis of 5 states. |0>, |1>, |2>, |3>, |4>.
    Models spin-2 particles like the graviton.
    """

    def __new__(cls, input_array):
        obj = super().__new__(cls, input_array)
        assert obj.is_at_least_nxn(n=5), "Gate must be at least a 5x5 matrix"
        assert obj.is_power_of_n_shape(n=5), "Gate shape must be a power of 5"
        return obj

    def num_qupents(self) -> int:
        """Return the number of qupents involved in the gate."""
        return int(np.log(self.shape[0]) / np.log(5))

    def is_multi_qupent(self) -> bool:
        """Check if the gate involves multiple qupents."""
        return self.num_qupents() > 1
